# $Header: /usr/home/johnpc/perl/uniud/RCS/uniud,v 1.14 2005/07/01 12:35:29 johnpc Exp $
# 
# NOTICE: This source contains UTF-8 unicode characters, but only in the
# comments. You can safely remove them if your editor barfs on them.

package Acme::Rot180;

use 5.008006;
use strict;
use warnings;

use Exporter qw(import);

our @EXPORT_OK = qw( upside_down ud_linelength ud_linelength_auto );

our $VERSION = '0.20';

use PerlIO;

our $has_termsize;
eval { require Term::Size; $has_termsize = 1; };

# turn ascii text upside-down, using unicode

my %updown = (
    ' ' => ' ',
    '!' => "\x{00a1}",          # ¡
    '"' => "\x{201e}",          # „
    '#' => '#',
    '$' => '$',
    '%' => '%',
    '&' => "\x{214b}",          # ⅋
    "'" => "\x{0375}",          # ͵
    '(' => ')',
    ')' => '(',
    '*' => '*',
    '+' => '+',
    ',' => "\x{2018}",          # ‘
    '-' => '-',
    '.' => "\x{02d9}",          # ˙
    '/' => '/',
    '0' => '0',
    '1' => "\x{002c}\x{20d3}",  # ,⃓ can be improved
    '2' => "\x{10f7}",          # ჷ
    '3' => "\x{03b5}",          # ε
    '4' => "\x{21c1}\x{20d3}",  # ⇁⃓ can be improved
    '5' => "\x{1515}",          # ᔕ or maybe just "S"
    '6' => '9',
    '7' => "\x{005f}\x{0338}",  # _̸
    '8' => '8',
    '9' => '6',
    ':' => ':',
    ';' => "\x{22c5}\x{0315}",  # ⋅̕ sloppy, should be improved
    '<' => '>',
    '=' => '=',
    '>' => '<',
    '?' => "\x{00bf}",          # ¿
    '@' => '@',                 # can be improved
    'A' => "\x{13cc}",          # Ꮜ
    'B' => "\x{03f4}",          # ϴ can be improved
    'C' => "\x{0186}",          # Ɔ
    'D' => 'p',                 # should be an uppercase D!!
    'E' => "\x{018e}",          # Ǝ
    'F' => "\x{2132}",          # Ⅎ
    'G' => "\x{2141}",          # ⅁
    'H' => 'H',
    'I' => 'I',
    'J' => "\x{017f}\x{0332}",  # ſ̲
    'K' => "\x{4e2c}",          #  丬
    'L' => "\x{2142}",          # ⅂
    'M' => "\x{019c}",          # Ɯ or maybe just "W"
    'N' => 'N',
    'O' => 'O',
    'P' => 'd',                 # should be uppercase P
    'Q' => "\x{053e}",          # Ծ can be improved
    'R' => "\x{0222}",          # Ȣ can be improved
    'S' => 'S',
    'T' => "\x{22a5}",          # ⊥
    'U' => "\x{144e}",          # ᑎ
    'V' => "\x{039b}",          # Λ
    'W' => 'M',
    'X' => 'X',
    'Y' => "\x{2144}",          # ⅄
    'Z' => 'Z',
    '[' => ']',
    '\\' => '\\',
    ']' => '[',
    '^' => "\x{203f}",          # ‿
    '_' => "\x{203e}",          # ‾
    '`' => "\x{0020}\x{0316}",  #  ̖
    'a' => "\x{0250}",          # ɐ
    'b' => 'q',
    'c' => "\x{0254}",          # ɔ
    'd' => 'p',
    'e' => "\x{01dd}",          # ǝ
    'f' => "\x{025f}",          # ɟ
    'g' => "\x{0253}",          # ɓ
    'h' => "\x{0265}",          # ɥ
#    'i' => "\x{0131}\x{0323}",  # ı̣
    'i' => "\x{1d09}",		# ᴉ
    'j' => "\x{017f}\x{0323}",  # ſ̣
    'k' => "\x{029e}",          # ʞ
    'l' => "\x{01ae}",          # Ʈ can be improved
    'm' => "\x{026f}",          # ɯ
    'n' => 'u',
    'o' => 'o',
    'p' => 'd',
    'q' => 'b',
    'r' => "\x{0279}",          # ɹ
    's' => 's',
    't' => "\x{0287}",          # ʇ
    'u' => 'n',
    'v' => "\x{028c}",          # ʌ
    'w' => "\x{028d}",          # ʍ
    'x' => 'x',
    'y' => "\x{028e}",          # ʎ
    'z' => 'z',
    '{' => '}',
    '|' => '|',
    '}' => '{',
    '~' => "\x{223c}",          # ∼
);
my $missing = "\x{fffd}";       # � replacement character

my $linelength;

sub ud_linelength {
    $linelength = $_[0] if @_;
    return $linelength;
}

sub ud_linelength_auto {
    return unless $has_termsize;
    my $termfd = $_[0] || *STDOUT{IO};
    $linelength = Term::Size::chars($termfd);
    ### Term::Size resets the :utf8 layer somehow
    binmode STDOUT, ":utf8";
    if ( !$linelength ) {
        warn "Cannot find terminal linelength\n";
        undef $linelength;
    }
    return $linelength;
}

sub upside_down {
    my $str = shift;
    my @input;
    my $maxlen = 0;
    for ( split /\n/, $str ) {
        ### expand tabs before calculating the length
        1 while s{^([^\t]*)\t}{$1 . " " x (8 - length($1) % 8)}e;
        $maxlen = length if length > $maxlen;
        push @input, $_;
    }
    my $len = defined $linelength ? $linelength :$maxlen;
    my $ret = '';
    for ( my $i = $#input; $i >= 0; $i-- ) {
        $ret .= _turnedstr($input[$i], $len) . "\n";
    }
    return $ret;
}

# turnedstr - handle turning one string, respecting linelength.
# also expands tabs to spaces
sub _turnedstr {
    my $str = shift;
    my $linelength = shift;
    my $turned = '';
    my $tlength = 0;

    for my $char ( $str =~ /(\X)/g ) {
        if ( exists $updown{$char} ) {
            my $t = $updown{$char};
            $t = $missing if !length($t);
            $turned = $t . $turned;
            $tlength++;
        }
        elsif ( $char eq "\t" ) {
            my $tablen = 8 - $tlength % 8;
            $turned = " " x $tablen . $turned;
            $tlength += $tablen;
        }
        elsif ( ord($char) >= 32 ) {
            ### other chars copied literally
            $turned = $char . $turned;
            $tlength++;
        }
    }
    if ( defined($linelength) and $tlength < $linelength ) {
        $turned = " " x ( $linelength - $tlength) . $turned;
    }

    return $turned;
}

1;

__END__

=head1 NAME

Acme::Rot180 - Perl extension to turn text upside-down using unicode
characters.

=head1 SYNOPSIS

  use Acme::Rot180 qw(upside_down);

  print upside_down("Just say YES");

  use Acme::Rot180 qw(ud_linelength ud_linelength_auto);

  ud_linelength_auth() or ud_linelength(72);

  print upside_down(<<BLOCK);
     This entire block is turned upside-down and right-aligned
     to your right margin...
  BLOCK

=head1 DESCRIPTION

=head2 EXPORT

None by default, but see SYNOPSIS for the exportable subroutines.

=head2 BUGS

This has no object interface, yet. But I might add one in the
future, while keeping the exportable functions present using
a singleton object internally.

The mapping is missing several good mappings to upside-down
uppercase characters like D, P, K. Also, the mapping for several
characters is sub-optimal, and uses combining characters for
some, which doesn't work everywhere.

Also some unicode fonts are not that complete resulting in dotted
rectangles somewhere, but that's hardly a bug in Acme::Rot180.

There might be methods in the future to select a mapping that is
optimized for a certain architecture (eg: optimized for Mac,
Linux, etc).

=head1 HISTORY

=over 8

=item 0.13

Bla

=item 0.14

packed into a tar file without a directory, so it would overwrite the
README in your current directory.

=item 0.20

Modularized

=back

=head1 SEE ALSO

http://www.xs4all.nl/~johnpc/uniud/

=head1 AUTHOR

Jan-Pieter Cornet, E<lt>johnpc@xs4all.nlE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Jan-Pieter Cornet

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.6 or,
at your option, any later version of Perl 5 you may have available.

=cut
